"""API client for FusionSolar Kiosk."""
import asyncio
import codecs
import json
import os
import time
from datetime import datetime

import aiohttp
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding


class FusionSolarRestApi:

    ENERGY_FLOW_PATH = (
        "/rest/pvms/web/station/v1/overview/energy-flow?"
        + "stationDn=%STATION%&_=%CURRENT_UTC_TIME%"
    )
    STATION_DETAIL_PATH = (
        "/rest/pvms/web/station/v1/overview/station-detail?"
        + "stationDn=%STATION%&_=%CURRENT_UTC_TIME%"
    )
    ENERGY_BALANCE = (
        "/rest/pvms/web/station/"
        + "v1/overview/energy-balance?stationDn=%STATION%&"
        + "timeDim=2&queryTime=%CURRENT_UTC_TIME%&timeZone=2&"
        + "timeZoneStr=Europe%2FBerlin&_=%CURRENT_UTC_TIME%"
    )
    DEVICE_SIGNALS = (
        "/rest/pvms/web/device/v1/device-real-kpi?"
        + "signalIds=10025&signalIds=10032&signalIds=10029&signalIds=10019&"
        + "signalIds=10022&signalIds=10006&signalIds=10020&signalIds=10021&"
        + "signalIds=10027&signalIds=10028&signalIds=21029&signalIds=10018&"
        + "signalIds=10008&signalIds=10009&signalIds=10010&signalIds=10011&"
        + "signalIds=10012&signalIds=10013&signalIds=10014&signalIds=10015&"
        + "signalIds=10016&signalIds=11002&signalIds=11005&signalIds=11008&"
        + "signalIds=11011&signalIds=11014&signalIds=11017&signalIds=11020&"
        + "signalIds=11023&signalIds=11026&signalIds=11029&signalIds=11032&"
        + "signalIds=11035&signalIds=11038&signalIds=11041&signalIds=11044&"
        + "signalIds=11047&signalIds=11050&signalIds=11053&signalIds=11056&"
        + "signalIds=11059&signalIds=11062&signalIds=11065&signalIds=11068&"
        + "signalIds=11071&signalIds=11001&signalIds=11004&signalIds=11007&"
        + "signalIds=11010&signalIds=11013&signalIds=11016&signalIds=11019&"
        + "signalIds=11022&signalIds=11025&signalIds=11028&signalIds=11031&"
        + "signalIds=11034&signalIds=11037&signalIds=11040&signalIds=11043&"
        + "signalIds=11046&signalIds=11049&signalIds=11052&signalIds=11055&"
        + "signalIds=11058&signalIds=11061&signalIds=11064&signalIds=11067&"
        + "signalIds=11070&signalIds=14001&signalIds=14002&signalIds=14003&"
        + "signalIds=14004&signalIds=14005&signalIds=14006&signalIds=14007&"
        + "signalIds=14008&signalIds=14009&signalIds=14010&signalIds=14011&"
        + "signalIds=14012&signalIds=14013&signalIds=14014&signalIds=14015&"
        + "signalIds=14016&signalIds=14017&signalIds=14018&signalIds=14019&"
        + "signalIds=14020&signalIds=14021&signalIds=14022&signalIds=14023&"
        + "signalIds=14024&signalIds=10047&signalIds=10051&"
        + "deviceDn=%INVERTER_STATION%&_=%CURRENT_UTC_TIME%"
    )

    def __init__(self, logger, username, password, region="region01eu5"):
        self.logger = logger
        self._region = region
        self._username = username
        self._password = password
        self._stationdetail = None
        self._energy_balance = None
        self._device_signals = None
        self._inverter_station = None
        self._inverter_output_power = 0
        self._from_grid = 0
        self._to_grid = 0
        self._electrical_load = 0
        self._battery_soc = None
        self._battery_battery_power = None
        self._battery_charge_capacity = None
        self._battery_discharge_capacity = None
        self._string_output_power = 0

    async def get_pubkey(self, session):
        # pubkey to get pubkey
        url = "https://" + self._region + ".fusionsolar.huawei.com/unisso/pubkey"
        async with session.get(url) as resp:
            self.logger.debug(f"response from {url}: {resp}")
            self._regionhost = resp.host
            j = await resp.json()
            return j

    async def validate_user(self, session):
        url = (
            "https://"
            + self._region
            + ".fusionsolar.huawei.com"
            + "/unisso/"
            + self._validate_user_ver
            + "/validateUser.action"
        )

        body = {
            "multiRegionName": "",
            "organizationName": "",
            "username": self._username,
            "password": self._password_encrypted
            if self._validate_user_ver == "v3"
            else self._password,
        }

        headers = {
            "accept": "application/json, text/javascript, */*; q=0.01",
            "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
            "content-type": "application/json",
            "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-requested-with": "XMLHttpRequest",
        }

        async with session.post(
            url,
            json=body,
            headers=headers,
            max_redirects=20,
            params={
                "service": "/unisess/v1/auth?service=%2Fnetecowebext%2Fhome%2Findex.html",
                "decision": 1,
                "actionErrors": 465,
                "timeStamp": self._timestamp,
                "nonce": codecs.encode(os.urandom(16), "hex").decode(),
            },
        ) as resp:
            self.logger.debug(f"response from {url}: {resp}")
            j = await resp.json()
            return j

    async def login_redirect(self, session, redurl):
        headers = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
            "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "same-site",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
        }
        url = "https://" + self._region + ".fusionsolar.huawei.com" + redurl
        async with session.get(url, max_redirects=20, headers=headers) as resp:
            return resp

    async def get_jsessionid(self, session):
        url = "https://eu5.fusionsolar.huawei.com/rest/pvms/web/demouser/v1/region-list"
        async with session.get(url) as resp:
            return resp

    async def login(self):
        try:
            async with aiohttp.ClientSession(
                trust_env=True, cookies={"locale": "en-us"}
            ) as session:
                resp = await self.get_jsessionid(session)

                fusion_pubkey_json = await self.get_pubkey(session)
                self._validate_user_ver = "v2"
                if fusion_pubkey_json["enableEncrypt"]:
                    self._validate_user_ver = "v3"
                    self.public_key = serialization.load_pem_public_key(
                        fusion_pubkey_json["pubKey"].encode("UTF-8"),
                        backend=default_backend(),
                    )
                    hex_pwd = self.public_key.encrypt(
                        self._password.encode("UTF-8"),
                        padding.OAEP(
                            mgf=padding.MGF1(algorithm=hashes.SHA384()),
                            algorithm=hashes.SHA384(),
                            label=None,
                        ),
                    )
                    self._password_encrypted = (
                        codecs.encode(hex_pwd, "base64").decode().replace("\n", "")
                        + fusion_pubkey_json["version"]
                    )
                    self._timestamp = fusion_pubkey_json["timeStamp"]

                # validateUser.action to get bspsessionid
                json_resp = await self.validate_user(session)
                redurl = json_resp["redirectURL"]

                if redurl:
                    resp = await self.login_redirect(session, redurl)

                url = "https://" + self._region + ".fusionsolar.huawei.com"
                async with session.get(url, max_redirects=20) as resp2:
                    self.logger.debug(f"response from {url}: {resp2}")
                    if resp2.status == 200:
                        self._cookies = session.cookie_jar
                    else:
                        self.logger.error(f"Failed to retrieve cookies: {resp2}")
                        return False

                headers = {
                    "accept": "application/json",
                    "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
                    "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
                    "sec-ch-ua-mobile": "?0",
                    "sec-ch-ua-platform": '"Chrome OS"',
                    "sec-fetch-dest": "empty",
                    "sec-fetch-mode": "cors",
                    "sec-fetch-site": "same-origin",
                    "x-requested-with": "XMLHttpRequest",
                }

                url = (
                    "https://"
                    + self._region
                    + ".fusionsolar.huawei.com/unisess/v1/auth/session"
                )
                async with session.get(url, headers=headers) as resp3:
                    self.logger.debug(f"response from {url}: {resp3}")
                    response = await resp3.json()
                    csrfToken = response["csrfToken"]

                headers = {
                    "accept": "application/json, text/javascript, */*; q=0.01",
                    "accept-language": "en-AT,en;q=0.9",
                    "content-type": "application/json",
                    "roarand": csrfToken,
                    "sec-ch-ua": '"Chromium";v="104", " Not A;Brand";v="99", "Google Chrome";v="104"',
                    "sec-ch-ua-mobile": "?0",
                    "sec-ch-ua-platform": '"Chrome OS"',
                    "sec-fetch-dest": "empty",
                    "sec-fetch-mode": "cors",
                    "sec-fetch-site": "same-origin",
                    "x-non-renewal-session": "true",
                    "x-requested-with": "XMLHttpRequest",
                    "x-timezone-offset": "120",
                    "Referer": "https://"
                    + self._region
                    + ".fusionsolar.huawei.com/pvmswebsite/assets/build/index.html",
                    "Referrer-Policy": "strict-origin-when-cross-origin",
                }
                body = {
                    "curPage": 1,
                    "pageSize": 10,
                    "gridConnectionTime": "",
                    "queryTime": round(time.time() * 1000),
                    "timeZone": 2,
                    "sortId": "createTime",
                    "sortDir": "DESC",
                    "locale": "en_US",
                }
                url = (
                    "https://"
                    + self._region
                    + ".fusionsolar.huawei.com/rest/pvms/web/station/v1/station/station-list"
                )
                async with session.post(
                    url,
                    json=body,
                    headers=headers,
                ) as resp3:
                    self.logger.debug(f"response from {url}: {resp3}")
                    response = await resp3.json()
                    self._stationname = response["data"]["list"][0]["dn"]

            return True

        except Exception as ex:
            self.logger.exception(f"Failed to get data: {ex}")
            return False

    async def _get_rest_data(self, path):
        current_utc_time = int(datetime.utcnow().timestamp() * 1000)
        path = path.replace("%STATION%", self._stationname).replace(
            "%CURRENT_UTC_TIME%", str(current_utc_time)
        )
        if self._inverter_station:
            path = path.replace("%INVERTER_STATION%", self._inverter_station)

        url = "https://" + self._region + ".fusionsolar.huawei.com" + path
        headers = {
            "accept": (
                "text/html,application/xhtml+xml,application/xml;q=0.9,"
                "image/avif,image/webp,image/apng,*/*;"
                "q=0.8,application/signed-exchange;v=b3;q=0.9,"
                "application/json, text/javascript, */*; q=0.01"
            ),
            "accept-language": (
                "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5"
            ),
            "cache-control": "max-age=0",
            "content-type": "application/json",
            "sec-ch-ua": (
                '" Not A;Brand";v="99", "Chromium";v="100", "Google Chrome";v="100"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
        }
        resp = await self._get(url, headers)
        if len(resp) == 0:
            # try again
            resp = await self._get(url, headers)
        return resp

    async def send_idle(self):
        url = (
            "https://"
            + self._region
            + ".fusionsolar.huawei.com/rest/plat/smapp/v1/idle"
        )

        headers = {
            "accept": "application/json",
            "accept-language": (
                "en-AT,en;q=0.9,de-AT;q=0.8," "de;q=0.7,en-GB;q=0.6,en-US;q=0.5"
            ),
            "content-type": "application/json",
            "sec-ch-ua": (
                '".Not/A)Brand";v="99", "Google Chrome";v="103", "Chromium";v="103"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"ChromeOS"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-non-renewal-active": "true",
            "x-non-renewal-session": "false",
            "x-requested-with": "XMLHttpRequest",
            "Referer": (
                f"https://{self._region}.fusionsolar.huawei.com"
                "/pvmswebsite/assets/build/index.html"
            ),
            "Referrer-Policy": "strict-origin-when-cross-origin",
        }
        await self._get(url, headers)

    async def _get(self, url, headers, max_retries=5):
        try:
            response = {}
            async with aiohttp.ClientSession(
                trust_env=True, headers=headers, cookie_jar=self._cookies
            ) as session:
                retry = 1
                while retry < max_retries:
                    retry += 1
                    async with session.get(url) as resp:
                        if resp.status != 200:
                            # wait a few seconds before login
                            self.logger.error(
                                f"Response from {url} was {resp.status}, try to relogin"
                            )
                            await asyncio.sleep(30)
                            await self.login()
                            continue
                        else:
                            break

                response = await resp.json()

            if "success" in response and response["success"] is True:
                if "data" in response:
                    return response["data"]
                return response

            return response
        except Exception as ex:
            self.logger.exception(f"Failed to get data from {url}: {ex}")
            return {}

    async def update(self):
        await self.update_station_detail()
        await self.update_energy_balance()
        await self.update_energy_flow()
        await self.update_device_signals()
        await self.send_idle()

    async def update_energy_flow(self):
        self._inverter_output_power = "-"
        self._from_grid = 0
        self._to_grid = 0
        self._electrical_load = "-"
        self._string_output_power = "-"
        self._battery_soc = None
        self._battery_battery_power = None
        self._battery_charge_capacity = None
        self._battery_discharge_capacity = None
        self._energyflowdata = None

        # https://region01eu5.fusionsolar.huawei.com/rest/pvms/web/station/v1/overview/energy-flow?stationDn=STATION&_
        result = await self._get_rest_data(FusionSolarRestApi.ENERGY_FLOW_PATH)
        self._energyflowdata = result
        if "flow" in result:
            for node in result["flow"]["nodes"]:
                if node["name"] == "neteco.pvms.devTypeLangKey.inverter":
                    if node["deviceTips"]["ACTIVE_POWER"] == "--":
                        self._inverter_output_power = 0
                    else:
                        self._inverter_output_power = float(
                            node["deviceTips"]["ACTIVE_POWER"]
                        )
                    self._inverter_station = node["devIds"][0]
                elif node["name"] == "neteco.pvms.KPI.kpiView.electricalLoad":
                    self._electrical_load = float(
                        node["description"]["value"].replace(" kW", "")
                    )
                elif node["name"] == "neteco.pvms.devTypeLangKey.energy_store":
                    self._battery_soc = float(node["deviceTips"]["SOC"])
                    self._battery_charge_capacity = float(
                        node["deviceTips"]["CHARGE_CAPACITY"]
                    )
                    self._battery_battery_power = float(
                        node["deviceTips"]["BATTERY_POWER"]
                    )
                    self._battery_discharge_capacity = float(
                        node["deviceTips"]["DISCHARGE_CAPACITY"]
                    )
                elif node["name"] == "neteco.pvms.devTypeLangKey.string":
                    self._string_output_power = float(
                        node["description"]["value"].replace(" kW", "")
                    )
            for node in result["flow"]["links"]:
                if "description" in node and "label" in node["description"]:
                    if (
                        node["description"]["label"]
                        == "neteco.pvms.energy.flow.buy.power"
                    ):
                        if node["description"]["value"] == "--":
                            self._to_grid = 0
                            self._from_grid = 0
                        else:
                            if node["fromNode"] == "3":
                                # from grid is fromNode 3
                                self._from_grid = float(
                                    node["description"]["value"].replace(" kW", "")
                                )
                                self._to_grid = 0
                            else:
                                # to grid is fromNode 2
                                self._to_grid = float(
                                    node["description"]["value"].replace(" kW", "")
                                )
                                self._from_grid = 0

    async def update_energy_balance(self):
        self._energy_balance = await self._get_rest_data(
            FusionSolarRestApi.ENERGY_BALANCE
        )

    async def update_device_signals(self):
        self._device_signals = await self._get_rest_data(
            FusionSolarRestApi.DEVICE_SIGNALS
        )
        if "signals" not in self._device_signals:
            await asyncio.sleep(10)
            self._device_signals = await self._get_rest_data(
                FusionSolarRestApi.DEVICE_SIGNALS
            )

    async def update_station_detail(self):
        # https://region01eu5.fusionsolar.huawei.com/rest/pvms/web/station/v1/overview/station-detail?stationDn=STATION&_=
        self._stationdetail = await self._get_rest_data(
            FusionSolarRestApi.STATION_DETAIL_PATH
        )

    @property
    def station(self):
        return self._stationdetail["dn"]

    @property
    def co2_saved(self):
        return self._stationdetail["co2"]

    @property
    def battery_soc(self):
        return self._battery_soc

    @property
    def battery_charge_capacity(self):
        return self._battery_charge_capacity

    @property
    def battery_power(self):
        return self._battery_battery_power

    @property
    def battery_discharge_capacity(self):
        return self._battery_discharge_capacity

    @property
    def total_lifetime_energy(self):
        return float(self._stationdetail["cumulativeEnergy"])

    @property
    def total_current_day_energy(self):
        return float(self._stationdetail["dailyEnergy"])

    @property
    def total_current_month_energy(self):
        return float(self._stationdetail["monthEnergy"])

    @property
    def total_current_year_energy(self):
        return float(self._stationdetail["yearEnergy"])

    @property
    def grid_connected_time(self):
        return self._stationdetail["gridConnectedTime"]

    @property
    def installed_capacity(self):
        return self._stationdetail["installedCapacity"]

    @property
    def daily_self_use_energy(self):
        if "selfUseNrg" in self._stationdetail["realNrgKpi"]["dailyNrg"]:
            return round(self._stationdetail["realNrgKpi"]["dailyNrg"]["selfUseNrg"], 2)
        return 0

    @property
    def daily_use_energy(self):
        if "useNrg" in self._stationdetail["realNrgKpi"]["dailyNrg"]:
            return round(self._stationdetail["realNrgKpi"]["dailyNrg"]["useNrg"], 2)
        return 0

    @property
    def daily_self_use_ratio(self):
        return self._energy_balance["selfUsePowerRatioByUse"]

    @property
    def daily_self_use_solar_ratio(self):
        return self._energy_balance["selfUsePowerRatioByProduct"]

    @property
    def inverter_output_power(self):
        return self._inverter_output_power

    @property
    def string_output_power(self):
        return self._string_output_power

    @property
    def electrical_load(self):
        return self._electrical_load

    @property
    def grid_power(self):
        if self._to_grid == 0 and self._from_grid == 0:
            return 0

        if self._to_grid > 0:
            return -(self._to_grid)

        return self._from_grid

    @property
    def to_grid_power(self):
        return self._to_grid

    @property
    def from_grid_power(self):
        return self._from_grid

    @property
    def string_details(self):
        string_details = {}
        if "signals" in self._device_signals:
            for signal in (11001, 11004):
                string_details[f"string_pv{int((signal-11001)/3+1)}"] = {
                    "voltage": self._device_signals["signals"][str(signal)]["value"],
                    "current": self._device_signals["signals"][str(signal + 1)][
                        "value"
                    ],
                }
        return string_details

    @property
    def data_string(self):
        return (
            f"Stationdetail Data:\n{self._stationdetail}\n\n"
            f"Energyflow Data:\n{self._energyflowdata}\n\n"
            f"Energybalance Data:\n{self._energy_balance}\n\n"
            f"Device Signals Data:\n{self._device_signals}\n"
        )



#!/usr/bin/env python3
"""Constants for the Tuya integration."""

TUYA_ENDPOINT = {
    "https://openapi.tuyaus.com": "America",
    "https://openapi.tuyacn.com": "China",
    "https://openapi.tuyaeu.com": "Europe",
    "https://openapi.tuyain.com": "India",
    "https://openapi-ueaz.tuyaus.com": "EasternAmerica",
    "https://openapi-weaz.tuyaeu.com": "WesternEurope",
}



# PyTuya Module
# -*- coding: utf-8 -*-
"""
This file was taken from
https://github.com/rospogrigio/localtuya/tree/master/custom_components/localtuya/pytuya

Python module to interface with Tuya WiFi smart devices.

Mostly derived from Shenzhen Xenon ESP8266MOD WiFi smart devices
E.g. https://wikidevi.com/wiki/Xenon_SM-PW701U

Author: clach04
Maintained by: postlund

For more information see https://github.com/clach04/python-tuya

Classes
   TuyaInterface(dev_id, address, local_key=None)
       dev_id (str): Device ID e.g. 01234567891234567890
       address (str): Device Network IP Address e.g. 10.0.1.99
       local_key (str, optional): The encryption key. Defaults to None.

Functions
   json = status()          # returns json payload
   set_version(version)     #  3.1 [default] or 3.3
   detect_available_dps()   # returns a list of available dps provided by the device
   update_dps(dps)          # sends update dps command
   add_dps_to_request(dp_index)  # adds dp_index to the list of dps used by the
                                  # device (to be queried in the payload)
   set_dp(on, dp_index)   # Set value of any dps index.


Credits
 * TuyaAPI https://github.com/codetheweb/tuyapi by codetheweb and blackrozes
   For protocol reverse engineering
 * PyTuya https://github.com/clach04/python-tuya by clach04
   The origin of this python module (now abandoned)
 * LocalTuya https://github.com/rospogrigio/localtuya-homeassistant by rospogrigio
   Updated pytuya to support devices with Device IDs of 22 characters
"""

import asyncio
import base64
import binascii
import json
import logging
import struct
import time
import weakref
from abc import ABC, abstractmethod
from collections import namedtuple
from hashlib import md5

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

version_tuple = (9, 0, 0)
version = version_string = __version__ = "%d.%d.%d" % version_tuple
__author__ = "postlund"

_LOGGER = logging.getLogger(__name__)

TuyaMessage = namedtuple("TuyaMessage", "seqno cmd retcode payload crc")

SET = "set"
STATUS = "status"
HEARTBEAT = "heartbeat"
UPDATEDPS = "updatedps"  # Request refresh of DPS

PROTOCOL_VERSION_BYTES_31 = b"3.1"
PROTOCOL_VERSION_BYTES_33 = b"3.3"

PROTOCOL_33_HEADER = PROTOCOL_VERSION_BYTES_33 + 12 * b"\x00"

MESSAGE_HEADER_FMT = ">4I"  # 4*uint32: prefix, seqno, cmd, length
MESSAGE_RECV_HEADER_FMT = ">5I"  # 4*uint32: prefix, seqno, cmd, length, retcode
MESSAGE_END_FMT = ">2I"  # 2*uint32: crc, suffix

PREFIX_VALUE = 0x000055AA
SUFFIX_VALUE = 0x0000AA55

HEARTBEAT_INTERVAL = 10

# DPS that are known to be safe to use with update_dps (0x12) command
UPDATE_DPS_WHITELIST = [18, 19, 20]  # Socket (Wi-Fi)

# This is intended to match requests.json payload at
# https://github.com/codetheweb/tuyapi :
# type_0a devices require the 0a command as the status request
# type_0d devices require the 0d command as the status request, and the list of
# dps used set to null in the request payload (see generate_payload method)

# prefix: # Next byte is command byte ("hexByte") some zero padding, then length
# of remaining payload, i.e. command + suffix (unclear if multiple bytes used for
# length, zero padding implies could be more than one byte)
PAYLOAD_DICT = {
    "type_0a": {
        STATUS: {"hexByte": 0x0A, "command": {"gwId": "", "devId": ""}},
        SET: {"hexByte": 0x07, "command": {"devId": "", "uid": "", "t": ""}},
        HEARTBEAT: {"hexByte": 0x09, "command": {}},
        UPDATEDPS: {"hexByte": 0x12, "command": {"dpId": [18, 19, 20]}},
    },
    "type_0d": {
        STATUS: {"hexByte": 0x0D, "command": {"devId": "", "uid": "", "t": ""}},
        SET: {"hexByte": 0x07, "command": {"devId": "", "uid": "", "t": ""}},
        HEARTBEAT: {"hexByte": 0x09, "command": {}},
        UPDATEDPS: {"hexByte": 0x12, "command": {"dpId": [18, 19, 20]}},
    },
}


class TuyaLoggingAdapter(logging.LoggerAdapter):
    """Adapter that adds device id to all log points."""

    def process(self, msg, kwargs):
        """Process log point and return output."""
        dev_id = self.extra["device_id"]
        return f"[{dev_id[0:3]}...{dev_id[-3:]}] {msg}", kwargs


class ContextualLogger:
    """Contextual logger adding device id to log points."""

    def __init__(self):
        """Initialize a new ContextualLogger."""
        self._logger = None

    def set_logger(self, logger, device_id):
        """Set base logger to use."""
        self._logger = TuyaLoggingAdapter(logger, {"device_id": device_id})

    def debug(self, msg, *args):
        """Debug level log."""
        return self._logger.log(logging.DEBUG, msg, *args)

    def info(self, msg, *args):
        """Info level log."""
        return self._logger.log(logging.INFO, msg, *args)

    def warning(self, msg, *args):
        """Warning method log."""
        return self._logger.log(logging.WARNING, msg, *args)

    def error(self, msg, *args):
        """Error level log."""
        return self._logger.log(logging.ERROR, msg, *args)

    def exception(self, msg, *args):
        """Exception level log."""
        return self._logger.exception(msg, *args)


def pack_message(msg):
    """Pack a TuyaMessage into bytes."""
    # Create full message excluding CRC and suffix
    buffer = (
        struct.pack(
            MESSAGE_HEADER_FMT,
            PREFIX_VALUE,
            msg.seqno,
            msg.cmd,
            len(msg.payload) + struct.calcsize(MESSAGE_END_FMT),
        )
        + msg.payload
    )

    # Calculate CRC, add it together with suffix
    buffer += struct.pack(MESSAGE_END_FMT, binascii.crc32(buffer), SUFFIX_VALUE)

    return buffer


def unpack_message(data):
    """Unpack bytes into a TuyaMessage."""
    header_len = struct.calcsize(MESSAGE_RECV_HEADER_FMT)
    end_len = struct.calcsize(MESSAGE_END_FMT)

    _, seqno, cmd, _, retcode = struct.unpack(
        MESSAGE_RECV_HEADER_FMT, data[:header_len]
    )
    payload = data[header_len:-end_len]
    crc, _ = struct.unpack(MESSAGE_END_FMT, data[-end_len:])
    return TuyaMessage(seqno, cmd, retcode, payload, crc)


class AESCipher:
    """Cipher module for Tuya communication."""

    def __init__(self, key):
        """Initialize a new AESCipher."""
        self.block_size = 16
        self.cipher = Cipher(algorithms.AES(key), modes.ECB(), default_backend())

    def encrypt(self, raw, use_base64=True):
        """Encrypt data to be sent to device."""
        encryptor = self.cipher.encryptor()
        crypted_text = encryptor.update(self._pad(raw)) + encryptor.finalize()
        return base64.b64encode(crypted_text) if use_base64 else crypted_text

    def decrypt(self, enc, use_base64=True):
        """Decrypt data from device."""
        if use_base64:
            enc = base64.b64decode(enc)

        decryptor = self.cipher.decryptor()
        return self._unpad(decryptor.update(enc) + decryptor.finalize()).decode()

    def _pad(self, data):
        padnum = self.block_size - len(data) % self.block_size
        return data + padnum * chr(padnum).encode()

    @staticmethod
    def _unpad(data):
        return data[: -ord(data[len(data) - 1 :])]


class MessageDispatcher(ContextualLogger):
    """Buffer and dispatcher for Tuya messages."""

    # Heartbeats always respond with sequence number 0, so they can't be waited for like
    # other messages. This is a hack to allow waiting for heartbeats.
    HEARTBEAT_SEQNO = -100

    def __init__(self, dev_id, listener):
        """Initialize a new MessageBuffer."""
        super().__init__()
        self.buffer = b""
        self.listeners = {}
        self.listener = listener
        self.set_logger(_LOGGER, dev_id)

    def abort(self):
        """Abort all waiting clients."""
        for key in self.listeners:
            sem = self.listeners[key]
            self.listeners[key] = None

            # TODO: Received data and semahore should be stored separately
            if isinstance(sem, asyncio.Semaphore):
                sem.release()

    async def wait_for(self, seqno, timeout=5):
        """Wait for response to a sequence number to be received and return it."""
        if seqno in self.listeners:
            raise Exception(f"listener exists for {seqno}")

        self.debug("Waiting for sequence number %d", seqno)
        self.listeners[seqno] = asyncio.Semaphore(0)
        try:
            await asyncio.wait_for(self.listeners[seqno].acquire(), timeout=timeout)
        except asyncio.TimeoutError:
            del self.listeners[seqno]
            raise

        return self.listeners.pop(seqno)

    def add_data(self, data):
        """Add new data to the buffer and try to parse messages."""
        self.buffer += data
        header_len = struct.calcsize(MESSAGE_RECV_HEADER_FMT)

        while self.buffer:
            # Check if enough data for measage header
            if len(self.buffer) < header_len:
                break

            # Parse header and check if enough data according to length in header
            _, seqno, cmd, length, retcode = struct.unpack_from(
                MESSAGE_RECV_HEADER_FMT, self.buffer
            )
            if len(self.buffer[header_len - 4 :]) < length:
                break

            # length includes payload length, retcode, crc and suffix
            if (retcode & 0xFFFFFF00) != 0:
                payload_start = header_len - 4
                payload_length = length - struct.calcsize(MESSAGE_END_FMT)
            else:
                payload_start = header_len
                payload_length = length - 4 - struct.calcsize(MESSAGE_END_FMT)
            payload = self.buffer[payload_start : payload_start + payload_length]

            crc, _ = struct.unpack_from(
                MESSAGE_END_FMT,
                self.buffer[payload_start + payload_length : payload_start + length],
            )

            self.buffer = self.buffer[header_len - 4 + length :]
            self._dispatch(TuyaMessage(seqno, cmd, retcode, payload, crc))

    def _dispatch(self, msg):
        """Dispatch a message to someone that is listening."""
        self.debug("Dispatching message %s", msg)
        if msg.seqno in self.listeners:
            self.debug("Dispatching sequence number %d", msg.seqno)
            sem = self.listeners[msg.seqno]
            self.listeners[msg.seqno] = msg
            sem.release()
        elif msg.cmd == 0x09:
            self.debug("Got heartbeat response")
            if self.HEARTBEAT_SEQNO in self.listeners:
                sem = self.listeners[self.HEARTBEAT_SEQNO]
                self.listeners[self.HEARTBEAT_SEQNO] = msg
                sem.release()
        elif msg.cmd == 0x12:
            self.debug("Got normal updatedps response")
        elif msg.cmd == 0x08:
            self.debug("Got status update")
            self.listener(msg)
        else:
            self.debug(
                "Got message type %d for unknown listener %d: %s",
                msg.cmd,
                msg.seqno,
                msg,
            )


class TuyaListener(ABC):
    """Listener interface for Tuya device changes."""

    @abstractmethod
    def status_updated(self, status):
        """Device updated status."""

    @abstractmethod
    def disconnected(self):
        """Device disconnected."""


class EmptyListener(TuyaListener):
    """Listener doing nothing."""

    def status_updated(self, status):
        """Device updated status."""

    def disconnected(self):
        """Device disconnected."""


class TuyaProtocol(asyncio.Protocol, ContextualLogger):
    """Implementation of the Tuya protocol."""

    def __init__(self, dev_id, local_key, protocol_version, on_connected, listener):
        """
        Initialize a new TuyaInterface.

        Args:
            dev_id (str): The device id.
            address (str): The network address.
            local_key (str, optional): The encryption key. Defaults to None.

        Attributes:
            port (int): The port to connect to.
        """
        super().__init__()
        self.loop = asyncio.get_running_loop()
        self.set_logger(_LOGGER, dev_id)
        self.id = dev_id
        self.local_key = local_key.encode("latin1")
        self.version = protocol_version
        self.dev_type = "type_0a"
        self.dps_to_request = {}
        self.cipher = AESCipher(self.local_key)
        self.seqno = 0
        self.transport = None
        self.listener = weakref.ref(listener)
        self.dispatcher = self._setup_dispatcher()
        self.on_connected = on_connected
        self.heartbeater = None
        self.dps_cache = {}

    def _setup_dispatcher(self):
        def _status_update(msg):
            decoded_message = self._decode_payload(msg.payload)
            if "dps" in decoded_message:
                self.dps_cache.update(decoded_message["dps"])

            listener = self.listener and self.listener()
            if listener is not None:
                listener.status_updated(self.dps_cache)

        return MessageDispatcher(self.id, _status_update)

    def connection_made(self, transport):
        """Did connect to the device."""

        async def heartbeat_loop():
            """Continuously send heart beat updates."""
            self.debug("Started heartbeat loop")
            while True:
                try:
                    await self.heartbeat()
                    await asyncio.sleep(HEARTBEAT_INTERVAL)
                except asyncio.CancelledError:
                    self.debug("Stopped heartbeat loop")
                    break
                except asyncio.TimeoutError:
                    self.debug("Heartbeat failed due to timeout, disconnecting")
                    break
                except Exception as ex:  # pylint: disable=broad-except
                    self.exception("Heartbeat failed (%s), disconnecting", ex)
                    break

            transport = self.transport
            self.transport = None
            transport.close()

        self.transport = transport
        self.on_connected.set_result(True)
        self.heartbeater = self.loop.create_task(heartbeat_loop())

    def data_received(self, data):
        """Received data from device."""
        self.dispatcher.add_data(data)

    def connection_lost(self, exc):
        """Disconnected from device."""
        self.debug("Connection lost: %s", exc)
        try:
            listener = self.listener and self.listener()
            if listener is not None:
                listener.disconnected()
        except Exception:  # pylint: disable=broad-except
            self.exception("Failed to call disconnected callback")

    async def close(self):
        """Close connection and abort all outstanding listeners."""
        self.debug("Closing connection")
        if self.heartbeater is not None:
            self.heartbeater.cancel()
            try:
                await self.heartbeater
            except asyncio.CancelledError:
                pass
            self.heartbeater = None
        if self.dispatcher is not None:
            self.dispatcher.abort()
            self.dispatcher = None
        if self.transport is not None:
            transport = self.transport
            self.transport = None
            transport.close()

    async def exchange(self, command, dps=None):
        """Send and receive a message, returning response from device."""
        self.debug(
            "Sending command %s (device type: %s)",
            command,
            self.dev_type,
        )
        payload = self._generate_payload(command, dps)
        dev_type = self.dev_type

        # Wait for special sequence number if heartbeat
        seqno = (
            MessageDispatcher.HEARTBEAT_SEQNO
            if command == HEARTBEAT
            else (self.seqno - 1)
        )

        self.transport.write(payload)
        msg = await self.dispatcher.wait_for(seqno)
        if msg is None:
            self.debug("Wait was aborted for seqno %d", seqno)
            return None

        # TODO: Verify stuff, e.g. CRC sequence number?
        payload = self._decode_payload(msg.payload)

        # Perform a new exchange (once) if we switched device type
        if dev_type != self.dev_type:
            self.debug(
                "Re-send %s due to device type change (%s -> %s)",
                command,
                dev_type,
                self.dev_type,
            )
            return await self.exchange(command, dps)
        return payload

    async def status(self):
        """Return device status."""
        status = await self.exchange(STATUS)
        if status and "dps" in status:
            self.dps_cache.update(status["dps"])
        return self.dps_cache

    async def heartbeat(self):
        """Send a heartbeat message."""
        return await self.exchange(HEARTBEAT)

    async def update_dps(self, dps=None):
        """
        Request device to update index.

        Args:
            dps([int]): list of dps to update, default=detected&whitelisted
        """
        if self.version == 3.3:
            if dps is None:
                if not self.dps_cache:
                    await self.detect_available_dps()
                if self.dps_cache:
                    dps = [int(dp) for dp in self.dps_cache]
                    # filter non whitelisted dps
                    dps = list(set(dps).intersection(set(UPDATE_DPS_WHITELIST)))
            self.debug("updatedps() entry (dps %s, dps_cache %s)", dps, self.dps_cache)
            payload = self._generate_payload(UPDATEDPS, dps)
            self.transport.write(payload)
        return True

    async def set_dp(self, value, dp_index):
        """
        Set value (may be any type: bool, int or string) of any dps index.

        Args:
            dp_index(int):   dps index to set
            value: new value for the dps index
        """
        return await self.exchange(SET, {str(dp_index): value})

    async def set_dps(self, dps):
        """Set values for a set of datapoints."""
        return await self.exchange(SET, dps)

    async def detect_available_dps(self):
        """Return which datapoints are supported by the device."""
        # type_0d devices need a sort of bruteforce querying in order to detect the
        # list of available dps experience shows that the dps available are usually
        # in the ranges [1-25] and [100-110] need to split the bruteforcing in
        # different steps due to request payload limitation (max. length = 255)
        self.dps_cache = {}
        ranges = [(2, 11), (11, 21), (21, 31), (100, 111)]

        for dps_range in ranges:
            # dps 1 must always be sent, otherwise it might fail in case no dps is found
            # in the requested range
            self.dps_to_request = {"1": None}
            self.add_dps_to_request(range(*dps_range))
            try:
                data = await self.status()
            except Exception as ex:
                self.exception("Failed to get status: %s", ex)
                raise
            if "dps" in data:
                self.dps_cache.update(data["dps"])

            if self.dev_type == "type_0a":
                return self.dps_cache
        self.debug("Detected dps: %s", self.dps_cache)
        return self.dps_cache

    def add_dps_to_request(self, dp_indicies):
        """Add a datapoint (DP) to be included in requests."""
        if isinstance(dp_indicies, int):
            self.dps_to_request[str(dp_indicies)] = None
        else:
            self.dps_to_request.update({str(index): None for index in dp_indicies})

    def _decode_payload(self, payload):
        if not payload:
            payload = "{}"
        elif payload.startswith(b"{"):
            pass
        elif payload.startswith(PROTOCOL_VERSION_BYTES_31):
            payload = payload[len(PROTOCOL_VERSION_BYTES_31) :]  # remove version header
            # remove (what I'm guessing, but not confirmed is) 16-bytes of MD5
            # hexdigest of payload
            payload = self.cipher.decrypt(payload[16:])
        elif self.version == 3.3:
            if self.dev_type != "type_0a" or payload.startswith(
                PROTOCOL_VERSION_BYTES_33
            ):
                payload = payload[len(PROTOCOL_33_HEADER) :]
            payload = self.cipher.decrypt(payload, False)

            if "data unvalid" in payload:
                self.dev_type = "type_0d"
                self.debug(
                    "switching to dev_type %s",
                    self.dev_type,
                )
                return None
        else:
            raise Exception(f"Unexpected payload={payload}")

        if not isinstance(payload, str):
            try:
                payload = payload.decode()
            except Exception:
                self.error(f"Failed to decode: {payload}")
                return {}
        self.debug("Decrypted payload: %s", payload)
        return json.loads(payload)

    def _generate_payload(self, command, data=None):
        """
        Generate the payload to send.

        Args:
            command(str): The type of command.
                This is one of the entries from payload_dict
            data(dict, optional): The data to be send.
                This is what will be passed via the 'dps' entry
        """
        cmd_data = PAYLOAD_DICT[self.dev_type][command]
        json_data = cmd_data["command"]
        command_hb = cmd_data["hexByte"]

        if "gwId" in json_data:
            json_data["gwId"] = self.id
        if "devId" in json_data:
            json_data["devId"] = self.id
        if "uid" in json_data:
            json_data["uid"] = self.id  # still use id, no separate uid
        if "t" in json_data:
            json_data["t"] = str(int(time.time()))

        if data is not None:
            if "dpId" in json_data:
                json_data["dpId"] = data
            else:
                json_data["dps"] = data
        elif command_hb == 0x0D:
            json_data["dps"] = self.dps_to_request

        payload = json.dumps(json_data).replace(" ", "").encode("utf-8")
        self.debug("Send payload: %s", payload)

        if self.version == 3.3:
            payload = self.cipher.encrypt(payload, False)
            if command_hb not in [0x0A, 0x12]:
                # add the 3.3 header
                payload = PROTOCOL_33_HEADER + payload
        elif command == SET:
            payload = self.cipher.encrypt(payload)
            to_hash = (
                b"data="
                + payload
                + b"||lpv="
                + PROTOCOL_VERSION_BYTES_31
                + b"||"
                + self.local_key
            )
            hasher = md5()
            hasher.update(to_hash)
            hexdigest = hasher.hexdigest()
            payload = (
                PROTOCOL_VERSION_BYTES_31
                + hexdigest[8:][:16].encode("latin1")
                + payload
            )

        msg = TuyaMessage(self.seqno, command_hb, 0, payload, 0)
        self.seqno += 1
        return pack_message(msg)

    def __repr__(self):
        """Return internal string representation of object."""
        return self.id


async def connect(
    address,
    device_id,
    local_key,
    protocol_version,
    listener=None,
    port=6668,
    timeout=5,
):
    """Connect to a device."""
    loop = asyncio.get_running_loop()
    on_connected = loop.create_future()
    _, protocol = await loop.create_connection(
        lambda: TuyaProtocol(
            device_id,
            local_key,
            protocol_version,
            on_connected,
            listener or EmptyListener(),
        ),
        address,
        port,
    )

    await asyncio.wait_for(on_connected, timeout=timeout)
    return protocol



import asyncio
import base64
import hashlib
import logging
import re
import socket
from asyncio import StreamReader, StreamWriter
from typing import Union

_LOGGER = logging.getLogger(__name__)

# We should use HTTP-link because wget don't support HTTPS and curl removed in
# lastest fw. But it's not a problem because we check md5

# original link http://pkg.musl.cc/socat/mipsel-linux-musln32/bin/socat
# original link https://busybox.net/downloads/binaries/1.21.1/busybox-mipsel
WGET = "(wget http://master.dl.sourceforge.net/project/mgl03/{0}?viasf=1 -O /data/{1} && chmod +x /data/{1})"

CHECK_FIRMWARE = "/data/busybox lsattr /data/firmware/firmware_ota.bin"
LOCK_FIRMWARE = "mkdir -p /data/firmware && touch /data/firmware/firmware_ota.bin && /data/busybox chattr +i /data/firmware/firmware_ota.bin"
UNLOCK_FIRMWARE = "/data/busybox chattr -i /data/firmware/firmware_ota.bin"

RUN_FTP = "/data/busybox tcpsvd -E 0.0.0.0 21 /data/busybox ftpd -w &"
# flash on another ports because running ZHA or z2m can breake process
RUN_ZIGBEE_FLASH = "/data/ser2net -C '8115:raw:60:/dev/ttyS2:115200 8DATABITS NONE 1STOPBIT' -C '8038:raw:60:/dev/ttyS2:38400 8DATABITS NONE 1STOPBIT'"

TAR_DATA = (
    b"tar -czOC /data basic_app basic_gw conf factory miio "
    b"mijia_automation silicon_zigbee_host zigbee zigbee_gw "
    b"ble_info miioconfig.db 2>/dev/null | base64\n"
)

MD5_BT = {
    "1.4.6_0012": "367bf0045d00c28f6bff8d4132b883de",
    "1.4.6_0043": "c4fa99797438f21d0ae4a6c855b720d2",
    "1.4.7_0115": "be4724fbc5223fcde60aff7f58ffea28",
    "1.4.7_0160": "9290241cd9f1892d2ba84074f07391d4",
    "1.5.0_0026": "9290241cd9f1892d2ba84074f07391d4",
    "1.5.0_0102": "9290241cd9f1892d2ba84074f07391d4",
}
MD5_BUSYBOX = "099137899ece96f311ac5ab554ea6fec"
# MD5_GW3 = 'c81b91816d4b9ad9bb271a5567e36ce9'  # alpha
MD5_SER2NET = "f27a481e54f94ea7613b461bda090b0f"


def sed(app: str, pattern: str, repl: str):
    """sed with extended regex and edit file in-place"""
    repl = (
        repl.replace("$", "\$")
        .replace("&", "\&")
        .replace("=", "\=")
        .replace("`", "\`")
        .replace('"', '\\"')
        .replace("\n", "\\n")
    )
    return f'sed -r "s={pattern}={repl}=" -i /tmp/daemon_{app}.sh'


# grep output to cloud and send it to MQTT, use awk because buffer
PATCH_MIIO_MQTT = sed(
    "miio",
    "^ +miio_client .+$",
    "miio_client -l 0 -o FILE_STORE -d $MIIO_PATH -n 128 | awk '/ot_agent_recv_handler_one.+(ble_event|properties_changed|heartbeat)/{print $0;fflush()}' | mosquitto_pub -t log/miio -l &",
)
# use patched silabs_ncp_bt from sourceforge and send stderr to MQTT
PATCH_BLETOOTH_MQTT = sed(
    "miio",
    "^ +silabs_ncp_bt .+$",
    "/data/silabs_ncp_bt /dev/ttyS1 $RESTORE 2>&1 >/dev/null | mosquitto_pub -t log/ble -l &",
)

PATCH_ZIGBEE_PARENTS = sed(
    "app",
    "^ +(Lumi_Z3GatewayHost_MQTT [^>]+).+$",
    "\\1-l 0 | mosquitto_pub -t log/z3 -l &",
)

# replace default Z3 to ser2net
PATCH_ZIGBEE_TCP1 = sed("app", "grep Lumi_Z3GatewayHost_MQTT", "grep ser2net")
PATCH_ZIGBEE_TCP2 = sed(
    "app",
    "^ +Lumi_Z3GatewayHost_MQTT .+$",
    "/data/ser2net -C '8888:raw:60:/dev/ttyS2:38400 8DATABITS NONE 1STOPBIT XONXOFF'",
)

# patch silabs_ncp_bt for storing data in tmp (memory)
PATCH_MEMORY_BLUETOOTH1 = "[ -d /tmp/miio ] || (cp -R /data/miio /tmp && cp /data/silabs_ncp_bt /tmp && sed -r 's=/data/=/tmp//=g' -i /tmp/silabs_ncp_bt)"
PATCH_MEMORY_BLUETOOTH2 = sed("miio", "^/data/silabs_ncp_bt", "/tmp/silabs_ncp_bt")
# every 5 min sync sqlite DB from memory to NAND if changed
PATCH_MEMORY_BLUETOOTH3 = sed(
    "miio",
    "^\tdo$",
    """\tdo
if [ ${#N} -eq 60 ]; then
  if [ "`md5sum /tmp/miio/mible_local.db|cut -d' ' -f1`" != "`md5sum /data/miio/mible_local.db|cut -d' ' -f1`" ]; then
    cp /tmp/miio/mible_local.db /data/miio
    echo "`date` bluetooth" >> /var/log/storage_sync.log
  fi; N=
fi; N=$N.
""",
)

# move zigbee DB to tmp (memory)
PATCH_MEMORY_ZIGBEE1 = "[ -d /tmp/zigbee_gw ] || cp -R /data/zigbee_gw /tmp"
PATCH_MEMORY_ZIGBEE2 = sed("app", "^ +zigbee_gw", "zigbee_gw -s /tmp/zigbee_gw/")
# every 5 min sync zigbee DB if device list changed
PATCH_MEMORY_ZIGBEE3 = sed(
    "app",
    "^\tdo$",
    """\tdo
if [ ${#N} -eq 60 ]; then
  if [ "`md5sum /tmp/zigbee_gw/device_properties.json|cut -d' ' -f1`" != "`md5sum /data/zigbee_gw/device_properties.json|cut -d' ' -f1`" ]; then
    cp /tmp/zigbee_gw/device_properties.json /data/zigbee_gw
    false | cp -i /tmp/zigbee_gw/*.json /data/zigbee_gw/ 2>/dev/null
    echo "`date` zigbee" >> /var/log/storage_sync.log
  fi; N=
fi; N=$N.
""",
)

# just for statistics
SAVE_SERIAL_STATS = "[ -f /tmp/serial ] || cp /proc/tty/driver/serial /tmp"

# if [ ! -e /tmp/bt_dont_need_startup ]; then
PATCH_DISABLE_BLUETOOTH = sed(
    "miio", "^ +if.+bt_dont_need_startup.+$", "if false; then"
)
PATCH_DISABLE_BUZZER1 = "[ -f /tmp/basic_gw ] || (cp /bin/basic_gw /tmp && sed -r 's=dev_query=xxx_query=' -i /tmp/basic_gw)"
PATCH_DISABLE_BUZZER2 = sed("miio", "^ +basic_gw", "/tmp/basic_gw")

SYNC_MEMORY_FILE = """[ "`md5sum /tmp/{0}|cut -d' ' -f1`" != "`md5sum /data/{0}|cut -d' ' -f1`" ] && cp /tmp/{0} /data/{0}"""

DB_BLUETOOTH = (
    "`ls -1t /data/miio/mible_local.db /tmp/miio/mible_local.db 2>/dev/null | sed q`"
)
DB_ZIGBEE = "`ls -1t /data/zigbee_gw/* /tmp/zigbee_gw/* 2>/dev/null | sed -r 's/[^/]+$/*.json/;q'`"


class TelnetShell:
    reader: StreamReader = None
    writer: StreamWriter = None

    ver: str = None

    async def connect(self, host: str, port=23) -> bool:
        try:
            coro = asyncio.open_connection(host, port, limit=1_000_000)
            self.reader, self.writer = await asyncio.wait_for(coro, 5)

            coro = self.reader.readuntil(b"login: ")
            await asyncio.wait_for(coro, 3)

            self.writer.write(b"admin\n")

            coro = self.reader.readuntil(b"\r\n# ")
            raw = await asyncio.wait_for(coro, timeout=3)
            if b"Password:" in raw:
                raise Exception("Telnet with password don't supported")

            self.writer.write(b"stty -echo\n")
            coro = self.reader.readuntil(b"# ")
            await asyncio.wait_for(coro, timeout=3)

            self.ver = await self.get_version()

            return True
        except:
            return False

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()

    async def exec(self, command: str, as_bytes=False) -> Union[str, bytes]:
        """Run command and return it result."""
        self.writer.write(command.encode() + b"\n")
        coro = self.reader.readuntil(b"# ")
        raw = await asyncio.wait_for(coro, timeout=10)
        return raw[:-2] if as_bytes else raw[:-2].decode()

    async def read_file(self, filename: str, as_base64=False):
        command = f"cat {filename}|base64" if as_base64 else f"cat {filename}"
        try:
            raw = await self.exec(command, as_bytes=True)
            # b"cat: can't open ..."
            return base64.b64decode(raw) if as_base64 else raw
        except:
            return None

    async def check_bin(self, filename: str, md5: str, url=None) -> bool:
        """Check binary md5 and download it if needed."""
        if md5 in await self.exec("md5sum /data/" + filename):
            return True
        elif url:
            await self.exec(WGET.format(url, filename))
            return await self.check_bin(filename, md5)
        else:
            return False

    async def check_zigbee_tcp(self):
        return await self.check_bin("ser2net", MD5_SER2NET, "bin/ser2net")

    async def run_zigbee_flash(self) -> bool:
        if not await self.check_zigbee_tcp():
            return False
        await self.exec("killall daemon_app.sh")
        await self.exec("killall Lumi_Z3GatewayHost_MQTT ser2net socat")
        await self.exec(RUN_ZIGBEE_FLASH)
        return True

    async def update_daemon_app(self, patches: list):
        await self.exec("killall daemon_app.sh")
        await self.exec(
            "killall Lumi_Z3GatewayHost_MQTT ser2net socat zigbee_gw; pkill -f log/z3"
        )

        if not patches:
            await self.exec(f"daemon_app.sh &")
            return

        await self.exec("cp /bin/daemon_app.sh /tmp")
        for patch in patches:
            await self.exec(patch)

        await self.exec(f"/tmp/daemon_app.sh {self.app_ps(patches)} &")

    async def update_daemon_miio(self, patches: list):
        """Run default daemon_miio if no patches. Or run patched daemon_miio
        with patches hash in process list.
        """
        await self.exec("killall daemon_miio.sh")
        await self.exec(
            "killall miio_client silabs_ncp_bt; killall -9 basic_gw; pkill -f 'log/ble|log/miio'"
        )

        if not patches:
            await self.exec(f"daemon_miio.sh &")
            return

        await self.exec("cp /bin/daemon_miio.sh /tmp")
        for patch in patches:
            await self.exec(patch)

        await self.exec(f"/tmp/daemon_miio.sh {self.miio_ps(patches)} &")

    async def memory_sync(self):
        resp = await self.exec(
            "ls --color=never /tmp/miio/mible_local.db /tmp/zigbee_gw/*"
        )
        for file in resp.split("\r\n"):
            if not file.startswith("/tmp/"):
                continue
            command = SYNC_MEMORY_FILE.format(file[5:])
            await self.exec(command)

    @staticmethod
    def app_ps(patches: list):
        if patches:
            return hashlib.md5("\n".join(patches).encode()).hexdigest()
        return "/bin/daemon_app.sh"

    @staticmethod
    def miio_ps(patches: list):
        if patches:
            return hashlib.md5("\n".join(patches).encode()).hexdigest()
        return "/bin/daemon_miio.sh"

    async def check_firmware_lock(self) -> bool:
        """Check if firmware update locked. And create empty file if needed."""
        resp = await self.exec(CHECK_FIRMWARE)
        return "-i-" in resp

    async def lock_firmware(self, enable: bool):
        if await self.check_bin("busybox", MD5_BUSYBOX, "bin/busybox"):
            command = LOCK_FIRMWARE if enable else UNLOCK_FIRMWARE
            await self.exec(command)

    def run_ftp(self):
        if self.check_bin("busybox", MD5_BUSYBOX, "bin/busybox"):
            self.exec(RUN_FTP)

    async def check_bt(self) -> bool:
        md5 = MD5_BT.get(self.ver)
        if not md5:
            return False
        # we use same name for bt utis so gw can kill it in case of update etc.
        return await self.check_bin("silabs_ncp_bt", md5, md5 + "/silabs_ncp_bt")

    async def run_public_mosquitto(self):
        await self.exec("killall mosquitto")
        await asyncio.sleep(0.5)
        await self.exec("mosquitto -d")
        await asyncio.sleep(0.5)
        # fix CPU 90% full time bug
        await self.exec("killall zigbee_gw")

    async def run_ntpd(self):
        await self.exec("ntpd -l")

    async def get_running_ps(self) -> str:
        return await self.exec("ps -ww | grep -v ' 0 SW'")

    async def tar_data(self):
        self.writer.write(TAR_DATA)
        coro = self.reader.readuntil(b"\r\n")
        await asyncio.wait_for(coro, timeout=3)  # skip command
        coro = self.reader.readuntil(b"# ")
        raw = await asyncio.wait_for(coro, timeout=10)
        return base64.b64decode(raw)

    async def get_version(self):
        raw = await self.read_file("/etc/rootfs_fw_info")
        m = re.search(r"version=([0-9._]+)", raw.decode())
        return m[1]

    async def get_token(self):
        raw = await self.read_file("/data/miio/device.token")
        return raw.rstrip().hex()

    async def get_did(self):
        raw = await self.read_file("/data/miio/device.conf")
        m = re.search(r"did=(\d+)", raw.decode())
        return m[1]

    async def get_wlan_mac(self) -> str:
        raw = await self.read_file("/sys/class/net/wlan0/address")
        return raw.decode().rstrip().upper()

    @property
    def mesh_group_table(self) -> str:
        return "mesh_group_v3" if self.ver >= "1.4.7_0160" else "mesh_group_v1"

    @property
    def mesh_device_table(self) -> str:
        return "mesh_device_v3" if self.ver >= "1.4.7_0160" else "mesh_device"


NTP_DELTA = 2208988800  # 1970-01-01 00:00:00
NTP_QUERY = b"\x1b" + 47 * b"\0"


def ntp_time(host: str) -> float:
    """Return server send time"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(2)
    try:
        sock.sendto(NTP_QUERY, (host, 123))
        raw = sock.recv(1024)

        integ = int.from_bytes(raw[-8:-4], "big")
        fract = int.from_bytes(raw[-4:], "big")
        return integ + float(fract) / 2 ** 32 - NTP_DELTA
    except:
        return 0
    finally:
        sock.close()


def check_port(host: str, port: int):
    """Check if gateway port open."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(2)
    try:
        return s.connect_ex((host, port)) == 0
    finally:
        s.close()



import asyncio
import time

from fhempy.lib import fhem

from .base import BaseDevice


class HTSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)


class SmokeSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)


class WaterLeakSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)


class MotionSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)
        self.reset_task = None
        self._attr_reset_motion = 120

    def update(self, data):
        asyncio.run_coroutine_threadsafe(self.async_update(data), self.loop).result()

    async def async_update(self, data):
        self.last_update = time.time()
        if data is None:
            return

        await fhem.readingsBeginUpdate(self.hash)
        # device is online
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        # update data
        for reading in data:
            if reading == "motion":
                await fhem.readingsBulkUpdate(self.hash, reading, str(data[reading]))
            elif reading == "added_device":
                pass
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading.replace(" ", "_"), str(data[reading])
                )
        await fhem.readingsEndUpdate(self.hash, 1)

        if "motion" in data and "voltage" in data:
            return

        if self.reset_task:
            self.reset_task.cancel()

        self.reset_task = self.create_async_task(self._set_no_motion())

    async def _set_no_motion(self):
        try:
            await asyncio.sleep(self._attr_reset_motion)
            await fhem.readingsSingleUpdate(self.hash, "motion", "0", 1)
        except asyncio.CancelledError:
            pass


class ContactSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)
